#
#  maze.rb
#  MazeWarper
#
#  Created by Mayur Pawashe on 8/30/08.
#  Copyright (c) 2008 zgcoder. All rights reserved.
#

require 'cell'

class Maze
  attr_reader :cells, :size
  attr_accessor :cube_radius
  
  def initialize(size, cube_radius)
    @cells = Array.new
    @size = size
    @cube_radius = cube_radius
    
    for row in (0 .. @size - 1)
      for column in (0 .. @size - 1)
        new_cell = Cell.new(column, row)
        @cells << new_cell
      end
    end
    
    make_neighbors_frontier(@cells[0])
    @cells[0].bottom_wall.passage = false
    
    until (frontier_maze_cells = frontier_cells()).empty?
      frontier_cell = frontier_maze_cells[rand(frontier_maze_cells.length)]
      in_neighbor_cells = in_neighbors(frontier_cell)
      
      unless in_neighbor_cells.empty?
        in_neighbor_cell = in_neighbor_cells[rand(in_neighbor_cells.length)]
        
        if frontier_cell.x < in_neighbor_cell.x
          frontier_cell.right_wall.passage = true
          in_neighbor_cell.left_wall.passage = true
        elsif frontier_cell.x > in_neighbor_cell.x
          frontier_cell.left_wall.passage = true
          in_neighbor_cell.right_wall.passage = true
        elsif frontier_cell.y < in_neighbor_cell.y
          frontier_cell.top_wall.passage = true
          in_neighbor_cell.bottom_wall.passage = true
        elsif frontier_cell.y > in_neighbor_cell.y
          frontier_cell.bottom_wall.passage = true
          in_neighbor_cell.top_wall.passage = true
        end
      end
      
      make_neighbors_frontier(frontier_cell)
    end
    
    @cells[@size * @size - 1].top_wall.passage = true
  end
  
  def last_cell
    @cells[@size * @size - 1]
  end
  
  def current_cell_of_player(player)
    self.cell_at(player.position.x, player.position.z.abs)
  end
  
  def cell_at(x, y)
    row = (y / (self.cube_radius * 2)).floor
    column = (x / (self.cube_radius * 2)).floor
    
    if @size * row + column < @cells.length
      return @cells[@size * row + column]
    else
      return nil
    end
  end
  
  #precondition: cell.x > 0 and cell.x < size()
  def left_neighbor(cell)
    index = @size * cell.y + (cell.x - 1)
    if index >= 0 and index < @cells.length
      return @cells[index]
    else
      return nil
    end
  end
  
  #precondition: cell.x >= 0 and cell.x < size() - 1
  def right_neighbor(cell)
    index = @size * cell.y + (cell.x + 1)
    if index >= 0 and index < @cells.length
      @cells[index]
    else
      return nil
    end
  end
  
  #precondition: cell.y > 0 and cell.y < size()
  def bottom_neighbor(cell)
    index = @size * (cell.y - 1) + cell.x
    if index >= 0 and index < @cells.length
      @cells[index]
    else
      return nil
    end
  end
  
  #precondition: cell.y >= 0 and cell.y < size() - 1
  def top_neighbor(cell)
    index = @size * (cell.y + 1) + cell.x
    if index > 0 and index < @cells.length
      @cells[index]
    else
      return nil
    end
  end
  
  private
  
  def neighbors(cell)
    neighbors = []
    
    if cell.x > 0
      neighbors << @cells[@size * cell.y + (cell.x - 1)]
    end
    
    if cell.x < @size - 1
      neighbors << @cells[@size * cell.y + (cell.x + 1)]
    end
    
    if cell.y > 0
      neighbors << @cells[@size * (cell.y - 1) + cell.x]
    end
    
    if cell.y < @size - 1
      neighbors << @cells[@size * (cell.y + 1) + cell.x]
    end
    
    neighbors
  end
  
  def make_neighbors_frontier(the_cell)
    the_cell.type = CELL_TYPE_IN
    for cell in neighbors(the_cell)
      cell.type = CELL_TYPE_FRONTIER if cell.type == CELL_TYPE_OUT
    end
  end
  
  def in_neighbors(cell)
    the_neighbors = []
    
    for cell in neighbors(cell)
      the_neighbors << cell if cell.type == CELL_TYPE_IN
    end
    
    the_neighbors
  end
  
  def frontier_cells
    frontier_cells = []
    
    for cell in @cells
      frontier_cells << cell if cell.type == CELL_TYPE_FRONTIER
    end
    
    frontier_cells
  end
end
